@@ -50,16 +50,19 @@ module Network.Wai.Logger (
   , showSockAddr
   ) where
 
-import Control.Concurrent (forkIO, threadDelay, killThread)
+import Control.Applicative ((<$>))
+import Control.AutoUpdate (mkAutoUpdate, defaultUpdateSettings, updateAction)
+import Control.Concurrent (MVar, newMVar, tryTakeMVar, putMVar)
 import Control.Exception (handle, SomeException(..), bracket)
 import Control.Monad (when, void)
 import Network.HTTP.Types (Status)
 import Network.Wai (Request)
-import System.IO (withFile, hFileSize, IOMode(..))
+import System.EasyFile (getFileSize)
 import System.Log.FastLogger
 
 import Network.Wai.Logger.Apache
 import Network.Wai.Logger.Date
+import Network.Wai.Logger.IORef
 import Network.Wai.Logger.IP (showSockAddr)
 
 ----------------------------------------------------------------
@@ -68,25 +71,16 @@ import Network.Wai.Logger.IP (showSockAddr)
 --   This 'ApacheLogger' writes log message to stdout.
 --   Each buffer (4K bytes) is flushed every second.
 withStdoutLogger :: (ApacheLogger -> IO a) -> IO a
-withStdoutLogger app = bracket setup teardown $ \(aplogger, _, _) ->
+withStdoutLogger app = bracket setup teardown $ \(aplogger, _) ->
     app aplogger
   where
     setup = do
-        (getter, updater) <- clockDateCacher
+        (getter, _updater) <- clockDateCacher
         apf <- initLogger FromFallback (LogStdout 4096) getter
         let aplogger = apacheLogger apf
-            flusher = logFlusher apf
             remover = logRemover apf
-            loop = do
-                threadDelay 1000000
-                updater
-                flusher
-                loop
-        t <- forkIO loop
-        return (aplogger, remover, t)
-    teardown (_, remover, t) = do
-        void remover
-        killThread t
+        return (aplogger, remover)
+    teardown (_, remover) = void remover
 
 ----------------------------------------------------------------
 
@@ -95,17 +89,11 @@ type ApacheLogger = Request -> Status -> Maybe Integer -> IO ()
 
 data ApacheLoggerActions = ApacheLoggerActions {
     apacheLogger :: ApacheLogger
-    -- | Flushing log messages in the buffers.
-    --   This is explicitly called from your program.
-    --   Probably, one second and 10 seconds is proper to stdout and
-    --   log files, respectively.
-    --   See the source code of 'withStdoutLogger'.
-  , logFlusher :: IO ()
-    -- | Rotating log files.
-    --   This is explicitly called from your program.
-    --   Probably, 10 seconds is proper.
+    -- | This is obsoleted. Rotation is done on-demand.
+    --   So, this is now an empty action.
   , logRotator :: IO ()
     -- | Removing resources relating Apache logger.
+    --   E.g. flushing and deallocating internal buffers.
   , logRemover :: IO ()
   }
 
@@ -117,6 +105,7 @@ data LogType = LogNone                     -- ^ No logging.
              | LogFile FileLogSpec BufSize -- ^ Logging to a file.
                                            --   'BufSize' is a buffer size
                                            --   for each capability.
+                                           --   File rotation is done on-demand.
              | LogCallback (LogStr -> IO ()) (IO ())
 
 ----------------------------------------------------------------
@@ -135,13 +124,11 @@ initLogger ipsrc (LogCallback cb flush) dateget = callbackLoggerInit ipsrc cb fl
 noLoggerInit :: IO ApacheLoggerActions
 noLoggerInit = return ApacheLoggerActions {
     apacheLogger = noLogger
-  , logFlusher = noFlusher
   , logRotator = noRotator
   , logRemover = noRemover
   }
   where
     noLogger _ _ _ = return ()
-    noFlusher = return ()
     noRotator = return ()
     noRemover = return ()
 
@@ -150,12 +137,10 @@ stdoutLoggerInit :: IPAddrSource -> BufSize -> DateCacheGetter
 stdoutLoggerInit ipsrc size dateget = do
     lgrset <- newStdoutLoggerSet size
     let logger = apache (pushLogStr lgrset) ipsrc dateget
-        flusher = flushLogStr lgrset
         noRotator = return ()
         remover = rmLoggerSet lgrset
     return ApacheLoggerActions {
         apacheLogger = logger
-      , logFlusher = flusher
       , logRotator = noRotator
       , logRemover = remover
       }
@@ -164,27 +149,34 @@ fileLoggerInit :: IPAddrSource -> FileLogSpec -> BufSize -> DateCacheGetter
                -> IO ApacheLoggerActions
 fileLoggerInit ipsrc spec size dateget = do
     lgrset <- newFileLoggerSet size $ log_file spec
-    let logger = apache (pushLogStr lgrset) ipsrc dateget
-        flusher = flushLogStr lgrset
-        rotator = logRotater lgrset spec
+    ref <- newIORef (0 :: Int)
+    mvar <- newMVar ()
+    let logger a b c = do
+            cnt <- decrease ref
+            apache (pushLogStr lgrset) ipsrc dateget a b c
+            when (cnt <= 0) $ tryRotate lgrset spec ref mvar
+        noRotator = return ()
         remover = rmLoggerSet lgrset
     return ApacheLoggerActions {
         apacheLogger = logger
-      , logFlusher = flusher
-      , logRotator = rotator
+      , logRotator = noRotator
       , logRemover = remover
       }
 
+decrease :: IORef Int -> IO Int
+decrease ref = atomicModifyIORef' ref (\x -> (x - 1, x - 1))
+
 callbackLoggerInit :: IPAddrSource -> (LogStr -> IO ()) -> IO () -> DateCacheGetter
                    -> IO ApacheLoggerActions
 callbackLoggerInit ipsrc cb flush dateget = do
-    let logger = apache cb ipsrc dateget
-        flusher = flush
+    flush' <- mkAutoUpdate defaultUpdateSettings
+        { updateAction = flush
+        }
+    let logger a b c = apache cb ipsrc dateget a b c >> flush'
         noRotator = return ()
         remover = return ()
     return ApacheLoggerActions {
         apacheLogger = logger
-      , logFlusher = flusher
       , logRotator = noRotator
       , logRemover = remover
       }
@@ -198,17 +190,35 @@ apache cb ipsrc dateget req st mlen = do
 
 ----------------------------------------------------------------
 
-logRotater :: LoggerSet -> FileLogSpec -> IO ()
-logRotater lgrset spec = do
-    over <- isOver
-    when over $ do
-        rotate spec
-        renewLoggerSet lgrset
+tryRotate :: LoggerSet -> FileLogSpec -> IORef Int -> MVar () -> IO ()
+tryRotate lgrset spec ref mvar = bracket lock unlock rotateFiles
   where
+    lock           = tryTakeMVar mvar
+    unlock Nothing = return ()
+    unlock _       = putMVar mvar ()
+    rotateFiles Nothing = return ()
+    rotateFiles _       = do
+        msiz <- getSize
+        case msiz of
+            -- A file is not available.
+            -- So, let's set a big value to the counter so that
+            -- this function is not called frequently.
+            Nothing -> writeIORef ref 1000000
+            Just siz
+                | siz > limit -> do
+                    rotate spec
+                    renewLoggerSet lgrset
+                    writeIORef ref $ estimate limit
+                | otherwise -> do
+                    writeIORef ref $ estimate (limit - siz)
     file = log_file spec
-    isOver = handle (\(SomeException _) -> return False) $ do
-        siz <- withFile file ReadMode hFileSize
-        return (siz > log_file_size spec)
+    limit = log_file_size spec
+    getSize = handle (\(SomeException _) -> return Nothing) $ do
+        -- The log file is locked by GHC.
+        -- We need to get its file size by the way not using locks.
+        Just . fromIntegral <$> getFileSize file
+    -- 200 is an ad-hoc value for the length of log line.
+    estimate x = fromInteger (x `div` 200)
 
 ----------------------------------------------------------------
 
